import math
import numpy as np
from functools import wraps, lru_cache
import hashlib
import cloudpickle
from pathlib import Path
import inspect
from mpi4py import MPI
from typing import Iterable
from pipcs import Config
import os
from collections import OrderedDict


class SummaryWriter():
    def __init__(self, config):
        config_dump = cloudpickle.dumps(config)
        hash = hashlib.sha1(config_dump).hexdigest()
        if isinstance(config, Iterable) and not isinstance(config, Config):
            config = config[0]
        # if config.writer.clear:
        #     rmtree(config.writer.log_dir, True)
        Path(config.writer.log_dir).mkdir(parents=True, exist_ok=True)
        with open(os.path.join(config.writer.log_dir, hash), 'wb') as f:
            f.write(config_dump)
        path = os.path.join(config.writer.log_dir, hash + '.log')
        self.columns = set()
        self.label_loc = 0
        self.f = open(path, 'w+')
        self.values = OrderedDict()

    def add(self, **kwargs):
        for k in kwargs:
            if k not in self.values:
                self.f.seek(0, os.SEEK_SET)
                contents = self.f.readlines()
                if not len(contents):
                    contents.append(k)
                else:
                    contents[0] = contents[0].rstrip() + f',{k}\n'
                self.f.seek(0, os.SEEK_SET)
                self.f.writelines(contents)
            self.values[k] = kwargs[k]

    def flush(self):
        for i, k in enumerate(self.values):
            if i:
                self.f.write(',')
            self.f.write(str(self.values[k]))
        self.f.write('\n')
        self.f.flush()

    def __exit__(self):
        self.f.close()


if MPI.COMM_WORLD.Get_rank() != 0:
    for name, fn in inspect.getmembers(SummaryWriter, inspect.isfunction):
        setattr(SummaryWriter, name, lambda *args, **kwargs: None)


def hook(func):
    func.before_register = []
    func.after_register = []
    def add_hook(after=True):
        def _add_hook(hook):
            if after:
                func.after_register.append(hook)
            else:
                func.before_register.append(hook)
        return _add_hook

    func.add_hook = add_hook

    @wraps(func)
    def wrapped(*args, **kwargs):
        for hook in func.before_register:
            hook(*args, **kwargs)
        ret = func(*args, **kwargs)
        for hook in func.after_register:
            hook(*args, **kwargs, ret=ret)
        return ret

    return wrapped


def calculate_cr(size):
    return np.exp(-0.5) / (np.log(size) * np.sqrt(2 * np.pi) + np.exp(-0.5))


@lru_cache(maxsize=1)
def _center_function(population_size):
    centers = np.arange(0, population_size, dtype=np.float32)
    centers = centers / (population_size - 1)
    centers -= 0.5
    centers *= 2.0
    return centers


def _compute_ranks(rewards):
    rewards = np.array(rewards)
    ranks = np.empty(rewards.size, dtype=int)
    ranks[rewards.argsort()] = np.arange(rewards.size)
    return ranks


def rank_transformation(rewards):
    ranks = _compute_ranks(rewards)
    values = _center_function(rewards.size)
    return values[ranks]



from typing import Callable, List, Optional, cast
from dataclasses import field
from itertools import tee
import copy
import os
import sys

from mpi4py import MPI
import gym
from pipcs import Config
import torch
from torch import nn
import numpy as np

from detorch import DE, Policy, default_config, hook, Strategy
import nes  # pip install nes-torch


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


class Agent(Policy):
    def __init__(self, input_size, hidden_layers, output_size, output_func,
                 hidden_act=None, output_act=None, bias=True):
        super().__init__()
        layers: List[nn.Module] = []
        layer_sizes = [input_size] + hidden_layers + [output_size]
        for idx, (layer1, layer2) in enumerate(pairwise(layer_sizes)):
            layers.append(nn.Linear(layer1, layer2, bias=bias))
            if hidden_act is not None:
                if idx < len(layer_sizes) - 2:
                    layers.append(hidden_act())
        if output_act is not None:
            layers.append(output_act())
        self.output_func = output_func
        self.seq = nn.Sequential(*layers)

    def rollout(self, env):
        done = False
        obs = env.reset()
        total_reward = 0
        while not done:
            obs_tensor = torch.from_numpy(obs).float()
            action = self.output_func(self.seq(obs_tensor))
            obs, reward, done, _ = env.step(action)
            total_reward += reward
        return total_reward


de_config = Config(default_config)
nes_config = Config(nes.default_config)


def make_env(id):
    return gym.make(id)


@de_config('environment')
@nes_config('environment')
class EnvironmentConfig():
    make_env = make_env
    id: str = 'LunarLander-v2'


@de_config('policy')
@nes_config('policy')
class PolicyConfig():
    policy = Agent
    env = DE.make_env(**de_config.environment.to_dict())
    input_size: int = env.observation_space.shape[0]
    if isinstance(env.action_space, gym.spaces.Discrete):
        output_size: int = env.action_space.n
        output_func: Callable[[int], int] = lambda x: x.argmax().item()
    else:
        output_size: int = env.action_space.shape[0]
        output_func: Callable[[int], int] = lambda x: x.detach().numpy()
    hidden_layers: List[int] = field(default_factory=lambda: [])
    # hidden_act: nn.Module = nn.ReLU
    bias: bool = True


@nes_config('policy')
class PolicyConfig():
    policy = type('Policy', (Agent, nes.Policy,), {})  # To supress assertion


@nes_config('optimizer')
class OptimizerConfig():
    lr = 0.02
    optim_type = torch.optim.Adam


@de_config('de', check=False)
@nes_config('nes', check=False)
class CommonConfig():
    n_rollout = 5
    seed = 123123


@nes_config('nes')
class NESConfig():
    n_step = 50
    population_size = 256
    l2_decay = 0.005
    sigma = 0.02

@de_config('de')
class DEConfig():
    n_step = 50
    population_size = 256
    differential_weight = (0.7, 1.0)
    crossover_probability = None
    strategy = Strategy.best1bin


if __name__ == '__main__':
    @nes.NES.__init__.add_hook()
    def init(self, *args, **kwargs):
        self.best_reward = -np.inf

    @DE.selection.add_hook()
    def after_selection(self, *args, **kwargs):
        print(f'Generation: {self.gen} Best Reward: {self.rewards[self.current_best]}')

    de = DE(de_config)

    @nes.NES.optimize.add_hook()
    def after_optimize(self, *args, **kwargs):
        reward = self.eval_policy(self.policy)
        if reward > self.best_reward:
            self.best_reward = reward
            self.best_policy = copy.deepcopy(self.policy)
        print(f'Generation: {self.gen+de.gen} Best Reward: {self.best_reward} Eval Reward: {reward}')

    de.train()

    nes = nes.NES(nes_config)

    nes.policy = nes.make_policy(**nes_config.policy.to_dict())
    nes.policy.load_state_dict(de.population[de.current_best].state_dict())
    nes.optim = nes.make_optimizer(nes.policy, **nes_config.optimizer.to_dict())

    nes.train()

    # Evaluate and Render
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    if comm.Get_rank() == 0:

        nes.env = gym.wrappers.Monitor(nes.env, './videos', force = True)

        setattr(nes, 'env.step', hook(nes.env.step.__func__))
        @nes.env.step.add_hook(after=False)
        def render(self, *args, **kwargs):
            self.render()

        print(f'Reward: {nes.eval_policy(nes.best_policy)}')



import os
import sys
import random
import copy
from abc import ABC, abstractmethod
from enum import IntEnum

import torch
from torch import nn
import numpy as np
from mpi4py import MPI
from pipcs import Config

from .utils import *


class Strategy(IntEnum):
    rand1bin = 0
    best1bin = 1
    rand2bin = 2
    best2bin = 3
    currenttobest1bin = 4
    randtobest1bin = 5
    scaledbest1bin = 6
    scaledrand1bin = 7
    rand1soft = 8
    best1soft = 9
    rand2soft = 10
    best2soft = 11
    currenttobest1soft = 12
    randtobest1soft = 13
    scaledbest1soft = 14
    scaledrand1soft = 15


class Policy(nn.Module, ABC):
    """Abstract subclass of nn.Module."""
    def __init__(self):
        super().__init__()

    @abstractmethod
    def rollout(self, env):
        """This function should be implemented by the user and
        it should evaluate the model in the given environment and
        return the total reward."""
        pass


class DE():
    """
    :ivar int gen: Current generation
    :ivar List[Policy] population
    :ivar gym.Env Gym environment
    """
    @hook
    def __init__(self, config: Config):
        config.check_config()
        self.config = config

        self.env = self.make_env(**config.environment.to_dict())

        np.random.seed(config.de.seed)
        random.seed(config.de.seed)
        torch.manual_seed(config.de.seed)
        torch.cuda.manual_seed(config.de.seed)
        torch.cuda.manual_seed_all(config.de.seed)
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True
        self.env.seed(config.de.seed)
        self.env.action_space.seed(config.de.seed)

        self.differential_weight = self.config.de.differential_weight

        if self.config.de.crossover_probability is None:
            policy = self.make_policy(**config.policy.to_dict())
            size = len(torch.nn.utils.parameters_to_vector(policy.parameters()))
            self.crossover_probability = calculate_cr(size)
        else:
            self.crossover_probability = self.config.de.crossover_probability

        self.rng = np.random.default_rng(config.de.seed)

        self.gen = 0

        self.population = []
        for _ in range(config.de.population_size):
            policy = self.make_policy(**config.policy.to_dict())
            self.population.append(policy)
        self.population = np.array(self.population)

    @staticmethod
    def make_env(make_env, **kwargs):
        """Helper function to create a gym environment."""
        return make_env(**kwargs)

    @staticmethod
    def make_policy(policy, **kwargs):
        """Helper function to create a policy."""
        assert issubclass(policy, Policy)
        return policy(**kwargs)

    @hook
    def eval_policy(self, policy):
        """Evaluate policy on the ``self.env`` for ``self.config.de.n_rollout times``"""
        total_reward = 0
        for _ in range(self.config.de.n_rollout):
            total_reward += policy.rollout(self.env)
        return total_reward / self.config.de.n_rollout

    def sample_parameters(self, size):
        idxs = self.rng.integers(0, len(self.population), size)
        samples = self.population[idxs]
        params = [torch.nn.utils.parameters_to_vector(sample.parameters()) for sample in samples]
        return params, idxs

    @hook
    def mutate(self, policy):
        """Mutate the given policy."""
        policy_params = torch.nn.utils.parameters_to_vector(policy.parameters())
        best_params = torch.nn.utils.parameters_to_vector(self.population[self.current_best].parameters())
        if self.config.de.strategy % 8 == Strategy.rand1bin:
            params, _ = self.sample_parameters(3)
            diff = params[1] - params[2]
            p0 = params[0]
        elif self.config.de.strategy % 8 == Strategy.best1bin:
            params, _ = self.sample_parameters(2)
            diff = params[0] - params[1]
            p0 = best_params
        elif self.config.de.strategy % 8 == Strategy.rand2bin:
            params, _ = self.sample_parameters(5)
            diff = params[1] - params[2] + params[3] - params[4]
            p0 = params[0]
        elif self.config.de.strategy % 8 == Strategy.best2bin:
            params, _ = self.sample_parameters(4)
            diff = params[0] - params[1] + params[2] - params[3]
            p0 = best_params
        elif self.config.de.strategy % 8 == Strategy.randtobest1bin:
            params, _ = self.sample_parameters(3)
            diff = best_params - params[0] + params[1] - params[2]
            p0 = params[0]
        elif self.config.de.strategy % 8 == Strategy.currenttobest1bin:
            params, _ = self.sample_parameters(2)
            diff = best_params - policy_params + params[0] - params[1]
            p0 = policy_params
        elif self.config.de.strategy % 8 == Strategy.scaledbest1bin:
            params, idxs = self.sample_parameters(2)
            rewards = rank_transformation(self.rewards)[idxs]
            distances = torch.stack(params) - policy_params
            diff = torch.from_numpy(rewards) @ distances
            p0 = best_params
        elif self.config.de.strategy % 8 == Strategy.scaledrand1bin:
            params, idxs = self.sample_parameters(3)
            rewards = rank_transformation(self.rewards)[idxs[1:]]
            distances = torch.stack(params[1:]) - policy_params
            diff = torch.from_numpy(rewards) @ distances
            p0 = params[0]
        else:
            raise NotImplementedError

        mutation = p0 + self.differential_weight * diff

        if self.config.de.strategy <= 7:
            cross = torch.rand(policy_params.shape) <= self.crossover_probability
            policy_params[cross] = mutation[cross]
        else:
            policy_params = self.crossover_probability * policy_params + (1.0 - self.crossover_probability) * mutation

        torch.nn.utils.vector_to_parameters(policy_params, policy.parameters())

    @hook
    def eval_population(self, population):
        comm = MPI.COMM_WORLD
        rewards = []
        reward_array = np.zeros(len(self.population), dtype=np.float32)
        batch = np.array_split(population, comm.Get_size())[comm.Get_rank()]
        for policy in batch:
            rewards.append(self.eval_policy(policy))
        comm.Allgatherv([np.array(rewards, dtype=np.float32), MPI.FLOAT], [reward_array, MPI.FLOAT])
        return reward_array

    @hook
    def generate_candidates(self):
        """Generate new candidates by mutating the population."""
        population = copy.deepcopy(self.population)
        for policy in population:
            self.mutate(policy)
        return population

    @hook
    def selection(self, population, rewards):
        """Select policies from the given population that are better than their parents."""
        comp = rewards >= self.rewards
        self.population[comp] = population[comp]
        self.rewards[comp] = rewards[comp]
        self.current_best = np.argmax(self.rewards)

    def train(self):
        torch.set_grad_enabled(False)
        rank = MPI.COMM_WORLD.Get_rank()
        if not rank == 0:
            f = open(os.devnull, 'w')
            sys.stdout = f

        self.rewards = self.eval_population(self.population)
        self.current_best = np.argmax(self.rewards)

        while self.gen < self.config.de.n_step:
            if isinstance(self.config.de.differential_weight, tuple):
                self.differential_weight = self.rng.uniform(*self.config.de.differential_weight)
            population = self.generate_candidates()
            rewards = self.eval_population(population)
            self.selection(population, rewards)
            self.gen += 1

        sys.stdout = sys.__stdout__
        torch.set_grad_enabled(True)



from typing import Callable, List, Optional
from dataclasses import field
from itertools import tee

import gym
from pipcs import Config
import torch
from torch import nn
import numpy as np

from detorch import DE, Policy, default_config, hook, Strategy


def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


class Agent(Policy):
    def __init__(self, input_size, hidden_layers, output_size, output_func,
                 hidden_act = None, output_act=None, bias=True):
        super().__init__()
        layers: List[nn.Module] = []
        layer_sizes = [input_size] + hidden_layers + [output_size]
        for idx, (layer1, layer2) in enumerate(pairwise(layer_sizes)):
            layers.append(nn.Linear(layer1, layer2, bias=bias))
            if hidden_act is not None:
                if idx < len(layer_sizes) - 2:
                    layers.append(hidden_act())
        if output_act is not None:
            layers.append(output_act())
        self.output_func = output_func
        self.seq = nn.Sequential(*layers)

    def rollout(self, env):
        done = False
        obs = env.reset()
        total_reward = 0
        while not done:
            obs_tensor = torch.from_numpy(obs).float()
            action = self.output_func(self.seq(obs_tensor))
            obs, reward, done, _ = env.step(action)
            total_reward += reward
        return total_reward


config = Config(default_config)


def make_env(id):
    return gym.make(id)


@config('environment')
class EnvironmentConfig():
    make_env = make_env
    id: str = 'LunarLander-v2'


@config('policy')
class PolicyConfig():
    policy = Agent
    env = DE.make_env(**config.environment.to_dict())
    input_size: int = env.observation_space.shape[0]
    if isinstance(env.action_space, gym.spaces.Discrete):
        output_size: int = env.action_space.n
        output_func: Callable[[int], int] = lambda x: x.argmax().item()
    else:
        output_size: int = env.action_space.shape[0]
        output_func: Callable[[int], int] = lambda x: x.detach().numpy()
    hidden_layers: List[int] = field(default_factory=lambda: [])
    # hidden_act: Optional[nn.Module] = nn.ReLU
    bias: bool = True


@config('de')
class DEConfig():
    n_rollout = 5
    n_step = 100
    population_size = 256
    differential_weight = (0.7, 1.0)
    crossover_probability = None
    strategy = Strategy.best1bin
    seed = 123123


if __name__ == '__main__':
    de = DE(config)

    @de.selection.add_hook()
    def after_selection(self, *args, **kwargs):
        print(f'Generation: {self.gen} Best Reward: {self.rewards[self.current_best]}')

    de.train()

    # Evaluate and render the best policy.
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    if comm.Get_rank() == 0:
        de.env = gym.wrappers.Monitor(de.env, './videos', force = True)
        setattr(de, 'env.step', hook(de.env.step.__func__))
        @de.env.step.add_hook(after=True)
        def render(self, *args, **kwargs):
            self.render()
        reward = de.eval_policy(de.population[de.current_best])
        print(f'Reward: {reward}')



import gym
from pipcs import Config
import torch
from torch import nn

from detorch import DE, Policy, default_config, Strategy


class Agent(Policy):
    def __init__(self):
        super().__init__()
        self.seq = nn.Sequential(
            nn.Linear(4, 8),
            nn.Linear(8, 2))

    def rollout(self, env):
        done = False
        obs = env.reset()
        total_reward = 0
        while not done:
            obs_tensor = torch.from_numpy(obs).float()
            action = self.seq(obs_tensor).argmax().item()
            obs, reward, done, _ = env.step(action)
            total_reward += reward
        return total_reward


config = Config(default_config)


@config('environment')
class EnvironmentConfig():
    make_env = gym.make
    id: str = 'CartPole-v1'


@config('policy')
class PolicyConfig():
    policy = Agent


@config('de')
class DEConfig():
    n_rollout = 2
    n_step = 5
    population_size = 256
    strategy = Strategy.best1bin
    seed = 123123



if __name__ == '__main__':
    de = DE(config)

    @de.selection.add_hook()
    def after_selection(self, *args, **kwargs):
        print(f'Generation: {self.gen} Best Reward: {self.rewards[self.current_best]}')


    de.train()
